@@ -9,8 +9,6 @@
 #![recursion_limit = "256"]
 #![cfg_attr(test, allow(dead_code))]
 #![warn(
-    missing_docs,
-    missing_debug_implementations,
     missing_copy_implementations,
     trivial_casts,
     trivial_numeric_casts,
@@ -27,6 +25,7 @@ mod crate_spec;
 mod dependency;
 mod errors;
 mod fetch;
+mod index;
 mod manifest;
 mod metadata;
 mod registry;
@@ -39,14 +38,13 @@ pub use dependency::PathSource;
 pub use dependency::RegistrySource;
 pub use dependency::Source;
 pub use errors::*;
-pub use fetch::{
-    get_compatible_dependency, get_latest_dependency, update_registry_index, RustVersion,
-};
-pub use manifest::{find, get_dep_version, set_dep_version, LocalManifest, Manifest};
+pub use fetch::{RustVersion, find_compatible_version, find_latest_version};
+pub use index::*;
+pub use manifest::{LocalManifest, Manifest, find, get_dep_version, set_dep_version};
 pub use metadata::manifest_from_pkgid;
 pub use registry::registry_url;
 pub use util::{
-    colorize_stderr, shell_note, shell_print, shell_status, shell_warn, shell_write_stderr,
-    shell_write_stdout, Color, ColorChoice,
+    Color, ColorChoice, colorize_stderr, shell_note, shell_print, shell_status, shell_warn,
+    shell_write_stderr, shell_write_stdout,
 };
-pub use version::{upgrade_requirement, VersionExt};
+pub use version::{VersionExt, upgrade_requirement};
