@@ -120,8 +120,10 @@
 #![cfg_attr(docsrs, feature(doc_auto_cfg))]
 
 use std::{
-    env, ffi, io, process,
-    sync::{Arc, Condvar, Mutex, MutexGuard},
+    env,
+    error::Error as StdError,
+    ffi, fmt, io, ops, process,
+    sync::{Arc, Mutex, MutexGuard, PoisonError},
 };
 
 use cfg_if::cfg_if;
@@ -140,6 +142,11 @@ cfg_if! {
     }
 }
 
+#[cfg(any(all(feature = "tokio", unix), not(any(unix, windows))))]
+mod async_client;
+#[cfg(any(all(feature = "tokio", unix), not(any(unix, windows))))]
+pub use async_client::AsyncAcquireClient;
+
 /// Command that can be accepted by this crate.
 pub trait Command {
     /// Inserts or updates an environment variable mapping.
@@ -255,6 +262,22 @@ where
     })
 }
 
+#[derive(Debug)]
+struct ClientInner {
+    inner: imp::Client,
+    #[cfg(unix)]
+    acitve_try_acquire_client_count: Mutex<usize>,
+}
+
+impl ClientInner {
+    #[cfg(unix)]
+    fn acitve_try_acquire_client_count(&self) -> MutexGuard<'_, usize> {
+        self.acitve_try_acquire_client_count
+            .lock()
+            .unwrap_or_else(PoisonError::into_inner)
+    }
+}
+
 /// A client of a jobserver
 ///
 /// This structure is the main type exposed by this library, and is where
@@ -269,9 +292,7 @@ where
 /// Note that a `Client` implements the `Clone` trait, and all instances of a
 /// `Client` refer to the same jobserver instance.
 #[derive(Clone, Debug)]
-pub struct Client {
-    inner: Arc<imp::Client>,
-}
+pub struct Client(Arc<ClientInner>);
 
 impl Client {
     /// Creates a new jobserver initialized with the given parallelism limit.
@@ -316,10 +337,12 @@ impl Client {
         }
     }
 
-    fn new_inner(client: imp::Client) -> Self {
-        Self {
-            inner: Arc::new(client),
-        }
+    fn new_inner(inner: imp::Client) -> Self {
+        Self(Arc::new(ClientInner {
+            inner,
+            #[cfg(unix)]
+            acitve_try_acquire_client_count: Mutex::default(),
+        }))
     }
 
     /// Attempts to connect to the jobserver specified in this process's
@@ -366,7 +389,37 @@ impl Client {
             .or_else(|| env::var_os("MAKEFLAGS"))
             .or_else(|| env::var_os("MFLAGS"))?;
 
-        imp::Client::open(var).map(|c| Client { inner: Arc::new(c) })
+        let var = {
+            cfg_if! {
+                if #[cfg(unix)] {
+                    std::os::unix::ffi::OsStrExt::as_bytes(var.as_os_str())
+                } else {
+                    var.to_str()?.as_bytes()
+                }
+            }
+        };
+        let makeflags = var.split(u8::is_ascii_whitespace);
+
+        // `--jobserver-auth=` is the only documented makeflags.
+        // `--jobserver-fds=` is actually an internal only makeflags, so we should
+        // always prefer `--jobserver-auth=`.
+        //
+        // Also, according to doc of makeflags, if there are multiple `--jobserver-auth=`
+        // the last one is used
+        if let Some(flag) = makeflags
+            .clone()
+            .filter_map(|s| s.strip_prefix(b"--jobserver-auth="))
+            .last()
+        {
+            imp::Client::open(flag)
+        } else {
+            imp::Client::open(
+                makeflags
+                    .filter_map(|s| s.strip_prefix(b"--jobserver-fds="))
+                    .last()?,
+            )
+        }
+        .map(Self::new_inner)
     }
 
     /// Acquires a token from this jobserver client.
@@ -386,7 +439,7 @@ impl Client {
     /// return immediately with the error. If an error is returned then a token
     /// was not acquired.
     pub fn acquire(&self) -> io::Result<Acquired> {
-        let data = self.inner.acquire()?;
+        let data = self.0.inner.acquire()?;
         Ok(Acquired::new(self, data))
     }
 
@@ -400,7 +453,7 @@ impl Client {
     ///
     /// Underlying errors from the ioctl will be passed up.
     pub fn available(&self) -> io::Result<usize> {
-        self.inner.available()
+        self.0.inner.available()
     }
 
     /// Configures a child process to have access to this client's jobserver as
@@ -451,11 +504,10 @@ impl Client {
         Cmd: Command,
         F: FnOnce(&mut Cmd) -> io::Result<R>,
     {
-        // Register one-time callback on unix to unset CLO_EXEC
-        // in child process.
-        self.inner.pre_run(&mut cmd);
+        #[cfg(unix)]
+        self.0.inner.pre_run(&mut cmd);
 
-        let arg = self.inner.string_arg();
+        let arg = self.0.inner.string_arg();
         // Older implementations of make use `--jobserver-fds` and newer
         // implementations use `--jobserver-auth`, pass both to try to catch
         // both implementations.
@@ -515,7 +567,7 @@ impl Client {
     {
         #[cfg(unix)]
         {
-            if let Some(path) = self.inner.get_fifo() {
+            if let Some(path) = self.0.inner.get_fifo() {
                 let path = path.as_os_str();
 
                 let prefix = "-j --jobserver-auth=";
@@ -533,102 +585,13 @@ impl Client {
         self.configure_and_run_inner(cmd, f, envs)
     }
 
-    /// Converts this `Client` into a helper thread to deal with a blocking
-    /// `acquire` function a little more easily.
-    ///
-    /// The fact that the `acquire` function on `Client` blocks isn't always
-    /// the easiest to work with. Typically you're using a jobserver to
-    /// manage running other events in parallel! This means that you need to
-    /// either (a) wait for an existing job to finish or (b) wait for a
-    /// new token to become available.
-    ///
-    /// Unfortunately the blocking in `acquire` happens at the implementation
-    /// layer of jobservers. On Unix this requires a blocking call to `read`
-    /// and on Windows this requires one of the `WaitFor*` functions. Both
-    /// of these situations aren't the easiest to deal with:
-    ///
-    /// * On Unix there's basically only one way to wake up a `read` early, and
-    ///   that's through a signal. This is what the `make` implementation
-    ///   itself uses, relying on `SIGCHLD` to wake up a blocking acquisition
-    ///   of a new job token. Unfortunately nonblocking I/O is not an option
-    ///   here, so it means that "waiting for one of two events" means that
-    ///   the latter event must generate a signal! This is not always the case
-    ///   on unix for all jobservers.
-    ///
-    /// * On Windows you'd have to basically use the `WaitForMultipleObjects`
-    ///   which means that you've got to canonicalize all your event sources
-    ///   into a `HANDLE` which also isn't the easiest thing to do
-    ///   unfortunately.
-    ///
-    /// This function essentially attempts to ease these limitations by
-    /// converting this `Client` into a helper thread spawned into this
-    /// process. The application can then request that the helper thread
-    /// acquires tokens and the provided closure will be invoked for each token
-    /// acquired.
-    ///
-    /// The intention is that this function can be used to translate the event
-    /// of a token acquisition into an arbitrary user-defined event.
-    ///
-    /// # Arguments
-    ///
-    /// This function will consume the `Client` provided to be transferred to
-    /// the helper thread that is spawned. Additionally a closure `f` is
-    /// provided to be invoked whenever a token is acquired.
-    ///
-    /// This closure is only invoked after calls to
-    /// `HelperThread::request_token` have been made and a token itself has
-    /// been acquired. If an error happens while acquiring the token then
-    /// an error will be yielded to the closure as well.
-    ///
-    /// # Return Value
-    ///
-    /// This function will return an instance of the `HelperThread` structure
-    /// which is used to manage the helper thread associated with this client.
-    /// Through the `HelperThread` you'll request that tokens are acquired.
-    /// When acquired, the closure provided here is invoked.
-    ///
-    /// When the `HelperThread` structure is returned it will be gracefully
-    /// torn down, and the calling thread will be blocked until the thread is
-    /// torn down (which should be prompt).
-    ///
-    /// # Errors
-    ///
-    /// This function may fail due to creation of the helper thread or
-    /// auxiliary I/O objects to manage the helper thread. In any of these
-    /// situations the error is propagated upwards.
-    ///
-    /// # Platform-specific behavior
-    ///
-    /// On Windows this function behaves pretty normally as expected, but on
-    /// Unix the implementation is... a little heinous. As mentioned above
-    /// we're forced into blocking I/O for token acquisition, namely a blocking
-    /// call to `read`. We must be able to unblock this, however, to tear down
-    /// the helper thread gracefully!
-    ///
-    /// Essentially what happens is that we'll send a signal to the helper
-    /// thread spawned and rely on `EINTR` being returned to wake up the helper
-    /// thread. This involves installing a global `SIGUSR1` handler that does
-    /// nothing along with sending signals to that thread. This may cause
-    /// odd behavior in some applications, so it's recommended to review and
-    /// test thoroughly before using this.
-    pub fn into_helper_thread<F>(self, f: F) -> io::Result<HelperThread>
-    where
-        F: FnMut(io::Result<Acquired>) + Send + 'static,
-    {
-        let state = Arc::new(HelperState::default());
-        Ok(HelperThread::new(
-            imp::spawn_helper(self, state.clone(), Box::new(f))?,
-            state,
-        ))
-    }
-
     /// Blocks the current thread until a token is acquired.
     ///
     /// This is the same as `acquire`, except that it doesn't return an RAII
     /// helper. If successful the process will need to guarantee that
     /// `release_raw` is called in the future.
     pub fn acquire_raw(&self) -> io::Result<()> {
-        self.inner.acquire()?;
+        self.0.inner.acquire()?;
         Ok(())
     }
 
@@ -638,9 +601,44 @@ impl Client {
     /// in some situations it could also be called to relinquish a process's
     /// implicit token temporarily which is then re-acquired later.
     pub fn release_raw(&self) -> io::Result<()> {
-        self.inner.release(None)?;
+        self.0.inner.release(None)?;
         Ok(())
     }
+
+    /// Get [`TryAcquireClient`], which supports non-blocking acquire.
+    ///
+    /// It would return `Err(IntoTryAcquireClientError::IncompatibleWithOlderMake)`
+    /// on unix if an annoymous pipe is used as jobserver.
+    ///
+    /// Once this function is called, on unix, `O_NONBLOCK` will bes et on jobserver
+    /// until the last instance of [`TryAcquireClient`] is dropped.
+    pub fn into_try_acquire_client(self) -> Result<TryAcquireClient, IntoTryAcquireClientError> {
+        #[cfg(unix)]
+        return {
+            // Construct `TryAcquireClient` here, in case `set_nonblocking`
+            // failed, its dtor would set it back to blocking.
+            let client = TryAcquireClient(self);
+
+            {
+                let mut active_try_acquire_client_count =
+                    client.0 .0.acitve_try_acquire_client_count();
+                *active_try_acquire_client_count += 1;
+
+                if *active_try_acquire_client_count == 1 {
+                    client.0 .0.inner.set_nonblocking()?;
+                }
+            }
+
+            if client.0 .0.inner.is_try_acquire_safe() {
+                Ok(client)
+            } else {
+                Err(IntoTryAcquireClientError::IncompatibleWithOlderMake(client))
+            }
+        };
+
+        #[cfg(not(unix))]
+        return Ok(TryAcquireClient(self));
+    }
 }
 
 /// An acquired token from a jobserver.
@@ -649,14 +647,14 @@ impl Client {
 /// otherwise represents the ability to spawn off another thread of work.
 #[derive(Debug)]
 pub struct Acquired {
-    client: Option<Arc<imp::Client>>,
+    client: Option<Arc<ClientInner>>,
     data: imp::Acquired,
 }
 
 impl Acquired {
     fn new(client: &Client, data: imp::Acquired) -> Self {
         Self {
-            client: Some(client.inner.clone()),
+            client: Some(client.0.clone()),
             data,
         }
     }
@@ -676,116 +674,152 @@ impl Acquired {
 impl Drop for Acquired {
     fn drop(&mut self) {
         if let Some(client) = self.client.take() {
-            drop(client.release(Some(&self.data)));
+            drop(client.inner.release(Some(&self.data)));
         }
     }
 }
 
+/// Possible errors for [`Client::into_try_acquire_client`]
 #[derive(Debug)]
-struct HelperThreadInner {
-    inner: imp::Helper,
-    state: Arc<HelperState>,
-}
-
-impl HelperThreadInner {
-    fn cleanup(self) {
-        // Flag that the producer half is done so the helper thread should exit
-        // quickly if it's waiting. Wake it up if it's actually waiting
-        self.state.lock().producer_done = true;
-        self.state.cvar.notify_one();
-
-        // ... and afterwards perform any thread cleanup logic
-        self.inner.join();
-    }
-}
-
-/// Structure returned from `Client::into_helper_thread` to manage the lifetime
-/// of the helper thread returned, see those associated docs for more info.
-#[derive(Debug)]
-#[repr(transparent)]
-pub struct HelperThread(ScopeGuard<HelperThreadInner, fn(HelperThreadInner)>);
-
-impl HelperThread {
-    fn new(inner: imp::Helper, state: Arc<HelperState>) -> Self {
-        Self(guard(
-            HelperThreadInner { inner, state },
-            HelperThreadInner::cleanup,
-        ))
-    }
-
-    /// Request that the helper thread acquires a token, eventually calling the
-    /// original closure with a token when it's available.
+pub enum IntoTryAcquireClientError {
+    /// The jobserver uses annoymous pipe, [`TryAcquireClient::try_acquire`]
+    /// requires setting `O_NONBLOCK`, since annoymous pipe is passed by fd,
+    /// it will affect all processes using the annoymous pipe,
+    /// which will break make < `4.4`.
     ///
-    /// For more information, see the docs on that function.
-    pub fn request_token(&self) {
-        // Indicate that there's one more request for a token and then wake up
-        // the helper thread if it's sleeping.
-        self.0.state.lock().requests += 1;
-        self.0.state.cvar.notify_one();
-    }
+    /// If you know that no make < `4.4` is spawned when [`TryAcquireClient`],
+    /// then you can simply unwrap this error and continue.
+    #[cfg(unix)]
+    IncompatibleWithOlderMake(TryAcquireClient),
+
+    /// Setting `O_NONBLOCK` failed.
+    #[cfg(unix)]
+    IoError(io::Error),
 }
 
-#[derive(Default, Debug)]
-struct HelperState {
-    lock: Mutex<HelperInner>,
-    cvar: Condvar,
-}
+impl fmt::Display for IntoTryAcquireClientError {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        match self {
+            #[cfg(unix)]
+            Self::IncompatibleWithOlderMake(_) => f.write_str(
+                r#"The jobserver uses annoymous pipe, `TryAcquireClient` will set `O_NONBLOCK`,
+since annoymous pipe is passed by fd, it will affect all processes using the annoymous pipe,
+which will break make < `4.4`."#,
+            ),
 
-#[derive(Default, Debug)]
-struct HelperInner {
-    requests: usize,
-    producer_done: bool,
-    consumer_done: bool,
-}
+            #[cfg(unix)]
+            Self::IoError(io_error) => write!(f, "io error: {}", io_error),
 
-impl HelperState {
-    fn lock(&self) -> MutexGuard<'_, HelperInner> {
-        self.lock.lock().unwrap_or_else(|e| e.into_inner())
-    }
-
-    /// Executes `f` for each request for a token, where `f` is expected to
-    /// block and then provide the original closure with a token once it's
-    /// acquired.
-    ///
-    /// This is an infinite loop until the helper thread is dropped, at which
-    /// point everything should get interrupted.
-    fn for_each_request(&self, mut f: impl FnMut(&HelperState)) {
-        let mut lock = self.lock();
-
-        // We only execute while we could receive requests, but as soon as
-        // that's `false` we're out of here.
-        while !lock.producer_done {
-            // If no one's requested a token then we wait for someone to
-            // request a token.
-            if lock.requests == 0 {
-                lock = self.cvar.wait(lock).unwrap_or_else(|e| e.into_inner());
-                continue;
-            }
-
-            // Consume the request for a token, and then actually acquire a
-            // token after unlocking our lock (not that acquisition happens in
-            // `f`). This ensures that we don't actually hold the lock if we
-            // wait for a long time for a token.
-            lock.requests -= 1;
-            drop(lock);
-            f(self);
-            lock = self.lock();
+            #[cfg(not(unix))]
+            _ => unreachable!(),
         }
-        lock.consumer_done = true;
-        self.cvar.notify_one();
-    }
-
-    fn producer_done(&self) -> bool {
-        self.lock().producer_done
     }
 }
 
-#[cfg(test)]
-mod tests {
-    #[test]
-    fn no_helper_deadlock() {
-        let x = crate::Client::new(32).unwrap();
-        let _y = x.clone();
-        std::mem::drop(x.into_helper_thread(|_| {}).unwrap());
+#[cfg(unix)]
+impl From<io::Error> for IntoTryAcquireClientError {
+    fn from(io_error: io::Error) -> Self {
+        Self::IoError(io_error)
+    }
+}
+
+impl StdError for IntoTryAcquireClientError {
+    fn source(&self) -> Option<&(dyn StdError + 'static)> {
+        match self {
+            #[cfg(unix)]
+            Self::IoError(io_error) => Some(io_error),
+            _ => None,
+        }
+    }
+}
+
+/// Extension of [`Client`] that supports non-blocking acquire.
+#[derive(Debug, derive_destructure2::destructure)]
+pub struct TryAcquireClient(Client);
+
+impl ops::Deref for TryAcquireClient {
+    type Target = Client;
+
+    fn deref(&self) -> &Self::Target {
+        &self.0
+    }
+}
+
+impl TryAcquireClient {
+    /// Similar to [`Client::acquire`], but returns `Ok(None)`
+    /// instead of bocking, if there is no token available.
+    pub fn try_acquire(&self) -> io::Result<Option<Acquired>> {
+        match self.0 .0.inner.try_acquire() {
+            Ok(Some(data)) => Ok(Some(Acquired::new(&self.0, data))),
+            Ok(None) => Ok(None),
+            Err(err) => Err(err),
+        }
+    }
+
+    /// Similar to [`Client::acquire_raw`], but returns `Ok(None)`
+    /// instead of blocking, if there is no token available.
+    pub fn try_acquire_raw(&self) -> io::Result<Option<()>> {
+        match self.0 .0.inner.try_acquire() {
+            Ok(Some(_)) => Ok(Some(())),
+            Ok(None) => Ok(None),
+            Err(err) => Err(err),
+        }
+    }
+
+    #[cfg(unix)]
+    fn cleanup(&self) -> io::Result<()> {
+        let mut active_try_acquire_client_count = self.0 .0.acitve_try_acquire_client_count();
+        *active_try_acquire_client_count -= 1;
+
+        if *active_try_acquire_client_count == 0 {
+            self.0 .0.inner.set_blocking()?;
+        }
+
+        Ok(())
+    }
+
+    /// Get back to [`Client`], return `Err` if clearing `O_NONBLOCK` fails.
+    pub fn into_inner(self) -> io::Result<Client> {
+        #[cfg(unix)]
+        self.cleanup()?;
+
+        Ok(self.destructure().0)
+    }
+}
+
+impl Drop for TryAcquireClient {
+    fn drop(&mut self) {
+        #[cfg(unix)]
+        let _ = self.cleanup();
+    }
+}
+
+#[cfg(unix)]
+impl std::os::unix::prelude::AsRawFd for TryAcquireClient {
+    fn as_raw_fd(&self) -> std::os::unix::prelude::RawFd {
+        self.0 .0.inner.get_read_fd()
+    }
+}
+
+#[cfg(any(unix, windows))]
+trait GenRandom {
+    fn new_random() -> io::Result<Self>
+    where
+        Self: Sized;
+}
+
+#[cfg(any(unix, windows))]
+impl GenRandom for u128 {
+    fn new_random() -> io::Result<Self> {
+        use std::mem::{transmute_copy, MaybeUninit};
+
+        const UNINIT_BYTE: MaybeUninit<u8> = MaybeUninit::<u8>::uninit();
+        let mut uninit_bytes = [UNINIT_BYTE; 16];
+
+        getrandom::fill_uninit(&mut uninit_bytes)?;
+
+        Ok(u128::from_ne_bytes(unsafe {
+            transmute_copy(&uninit_bytes)
+        }))
     }
 }
